import os
import logging
import json
from sortedcontainers import SortedDict
from collections import Counter

from elasticsearch_dsl import Search, Q
from nltk import download
from tqdm import tqdm
from nltk.stem.wordnet import WordNetLemmatizer

from ._bonn import FfModel
from .category_manager import CategoryManager
from .taxonomy import get_taxonomy, taxonomy_to_categories, categories_to_classifier_bow

try:
    from elasticsearch2 import Elasticsearch as Elasticsearch2
except ImportError:
    Elasticsearch2 = None

ELASTICSEARCH = None
try:
    from elasticsearch import Elasticsearch
    ELASTICSEARCH = Elasticsearch
except ImportError as exc:
    if not Elasticsearch2:
        raise RuntimeError("elasticsearch or elasticsearch2 must be installed") from exc

DEFAULT_TAXONOMY_LOCATION = "/app/test_data/taxonomy.json"


def get_elasticsearch(settings):
    elasticsearch_version = int(settings.get(
        "ELASTICSEARCH_MAJOR_VERSION",
        os.getenv("ELASTICSEARCH_MAJOR_VERSION", "7")
    ))

    elasticsearch = ELASTICSEARCH
    if elasticsearch_version == 2:
        if Elasticsearch2:
            elasticsearch = Elasticsearch2
        else:
            raise RuntimeError(
                "To use Elasticsearch 2.x, you must have elasticsearch_dsl~=2.0 "
                "and elasticsearch2 installed"
            )

    if elasticsearch is None:
        raise RuntimeError(
            "You must have a version of elasticsearch or elasticsearch2 installed"
        )

    return elasticsearch

def get_datasets(cm, classifier_bow, settings):
    ltzr = WordNetLemmatizer()
    host = settings.get(
        "ELASTICSEARCH_HOST",
        os.getenv("ELASTICSEARCH_HOST", "http://elasticsearch-master:9200"),
    )
    elasticsearch_index = settings.get(
        "ELASTICSEARCH_INDEX",
        os.getenv("ELASTICSEARCH_INDEX", "ons")
    )
    classifier_bow_vec = {
        k: [cm._model[w[1]] for w in words] for k, words in classifier_bow.items()
    }
    datasets = {}
    # results_df = pd.DataFrame((d.to_dict() for d in s.scan()))
    # /businesseconomy../business/activitiespeopel/123745
    elasticsearch = get_elasticsearch(settings)
    client = elasticsearch([host])

    s = Search(using=client, index=elasticsearch_index).filter(
        "bool", must=[Q("exists", field="title")]
    )
    expecting = s.count()
    size = 50
    s = s.params(size=size)
    all_words = Counter()
    with tqdm(total=expecting) as pbar:
        for hit in s.scan():
            try:
                title = hit.title
                datasets[title] = {
                    "category": tuple(hit.uri.split("/")[1:4]),
                    "text": f"{title} {hit.meta_description}",
                }
                cat = datasets[title]["category"]
                if cat not in classifier_bow_vec and cat[:-1] in classifier_bow_vec:
                    cat = cat[:-1]
                    datasets[title]["category"] = cat

                datasets[title]["bow"] = cm.closest(
                    datasets[title]["text"], cat, classifier_bow_vec
                )
                document = (
                    title
                    + " "
                    + datasets[title]["text"]
                )
                all_words.update(
                    {
                        ltzr.lemmatize(v)
                        for v in set(sum(cm.strip_document(document), []))
                    }
                )
                datasets[title]["bow"] = cm.closest(
                    document, cat, classifier_bow_vec
                )
            except AttributeError:
                pass
            pbar.update(1)

    all_words = {w: v for w, v in all_words.items() if v > 100}
    return datasets, all_words


def discover_terms(datasets, classifier_bow, settings):
    discovered_terms = {}
    # could do with lemmatizing
    for ds in datasets.values():
        if ds["category"][0:2] not in discovered_terms:
            discovered_terms[ds["category"][0:2]] = Counter()
        discovered_terms[ds["category"][0:2]].update(set(ds["bow"]))
        if ds["category"] not in discovered_terms:
            discovered_terms[ds["category"]] = Counter()
        discovered_terms[ds["category"]].update(set(ds["bow"]))

    appearance_threshold = settings.get("APPEARANCE_THRESHOLD", 5)
    upper_appearance_threshold = settings.get("UPPER_APPEARANCE_THRESHOLD", 10)
    discovered_terms = {
        k: [
            w
            for w, c in count.items()
            if c > (appearance_threshold if len(k) > 2 else upper_appearance_threshold)
        ]
        for k, count in discovered_terms.items()
    }
    for key, terms in classifier_bow.items():
        if key in discovered_terms:
            terms += [("WSSC", w) for w in discovered_terms[key]]
        if key[0:2] in discovered_terms:
            terms += [("WC", w) for w in discovered_terms[key[0:2]]]


def save_to_cache(cache_target, all_words, classifier_bow, settings):
    cached = {
        "config": dict(settings),
        "all-words": all_words,
        "classifier-bow": [
            ["|".join(key), [[c, v] for c, v in terms]]
            for key, terms in classifier_bow.items()
        ],
    }
    try:
        with open(cache_target, "w") as cache_f:
            json.dump(cached, cache_f)
    except OSError:
        logging.warning("Could not write cache to target %s", cache_target)


def append_discovered_terms_from_elasticsearch(cm, classifier_bow, settings):
    datasets, all_words = get_datasets(cm, classifier_bow, settings)
    discover_terms(datasets, classifier_bow, settings)
    cm.set_all_words(all_words)
    return all_words


def load(model_file, settings):
    model = FfModel(model_file)
    # Import and download stopwords from NLTK.
    download("stopwords")  # Download stopwords list.
    download("omw-1.4")  # Download lemma list.
    download("wordnet")  # Download lemma list.

    category_manager = CategoryManager(model, settings)
    taxonomy_location = settings.get(
        "TAXONOMY_LOCATION", DEFAULT_TAXONOMY_LOCATION
    )

    taxonomy = get_taxonomy(taxonomy_location)
    categories = taxonomy_to_categories(taxonomy)

    cache_target = settings.get("CACHE_TARGET", None)
    rebuild_cache = settings.get("REBUILD_CACHE", False)

    if cache_target and os.path.isfile(cache_target) and not rebuild_cache:
        with open(cache_target, "r") as cache_f:
            cached = json.load(cache_f)
        classifier_bow = SortedDict(
            {
                tuple(key.split("|")): tuple((c, v) for c, v in terms)
                for key, terms in cached["classifier-bow"]
            }
        )
        category_manager.set_all_words(cached["all-words"])
    else:
        classifier_bow = categories_to_classifier_bow(
            category_manager.strip_document, categories
        )
        all_words = append_discovered_terms_from_elasticsearch(
            category_manager, classifier_bow, settings
        )
        if cache_target and (not os.path.exists(cache_target) or rebuild_cache):
            save_to_cache(cache_target, all_words, classifier_bow, settings)

    category_manager.add_categories_from_bow("onyxcats", classifier_bow)

    return category_manager

if __name__ == "__main__":
    from .settings import settings
    load(settings["MODEL_FILE"], settings)



import numpy as np


def cosine_similarities(left, rights):
    return np.array(
        [np.dot(left / np.linalg.norm(left), v / np.linalg.norm(v)) for v in rights],
        dtype=np.float32,
    )


def _get_snr(value, scored_list_len, scored_list_sum):
    return value * (scored_list_len - 1) / (scored_list_sum - value)


def filter_by_snr(scored_list, snr):
    scored_list_sum = sum(s for s, _ in scored_list)
    scored_list_len = len(scored_list)
    return [
        (score, item)
        for score, item in scored_list
        if _get_snr(score, scored_list_len, scored_list_sum) > snr
    ]



import os
from bonn.settings import settings

def test_load_settings():
    assert settings.load_dotenv is True
    assert settings.ENVVAR_PREFIX_FOR_DYNACONF == 'BONN'

    os.environ["BONN_STOPWORDS_LANGUAGE"] = "welsh"
    settings.reload()
    assert settings.STOPWORDS_LANGUAGE == "welsh"



import pytest
from nltk import download
from pathlib import Path
from bonn.extract import CategoryManager, FfModel
from dynaconf import Dynaconf

def get_test_data(datafile):
    return Path(__file__).parent / "data" / datafile

@pytest.fixture(scope='module')
def model():
    model = FfModel(str(get_test_data("wiki.en.fifu")))
    # Import and download stopwords from NLTK.
    download("stopwords")  # Download stopwords list.
    download("omw-1.4")  # Download lemma list.
    download("wordnet")  # Download lemma list.
    return model



def make_category_manager(model, settings):
    category_manager = CategoryManager(model, settings)
    return category_manager

@pytest.fixture
def category_manager(model, settings):
    return make_category_manager(model, settings)

@pytest.fixture
def settings():
    settings = Dynaconf()
    settings.STOPWORDS_LANGUAGE = "english"
    return settings

def test_can_create_category_manager(category_manager):
    pass

def test_can_add_categories(category_manager):
    category_manager.add_categories_from_bow(
        "cities",
        {
            "manchester": (("C", "manchester"),),
            "cardiff": (("C", "cardiff"),),
            "ealing": (("C", "ealing"),),
            "underground": (("C", "underground"),),
        }
    )
    print(category_manager.test("london", "cities"))



from sortedcontainers import SortedDict
import json


def get_taxonomy(taxonomy_location):
    with open(taxonomy_location, "r") as f:
        taxonomy = json.load(f)
    return taxonomy


def categories_to_classifier_bow(strip_document, categories):
    classifier_categories = {
        catf: {
            subcatf: {
                subsubcatf: sum(
                    [
                        [("C", v) for w in strip_document(cat) for v in w]
                        + [("SC", v) for w in strip_document(subcat) for v in w]
                        + [("SSC", v) for w in strip_document(subsubcat) for v in w]
                    ],
                    [],
                )
                for subsubcatf, subsubcat in subcat_sub
            }
            if subcat_sub
            else {
                None: sum(
                    [
                        [("C", v) for w in strip_document(cat) for v in w]
                        + [("SC", v) for w in strip_document(subcat) for v in w]
                    ],
                    [],
                )
            }
            for (subcatf, subcat), subcat_sub in cat_sub.items()
        }
        for (catf, cat), cat_sub in categories.items()
    }
    classifier_bow = SortedDict(
        sum(
            [
                [
                    [(k, k2, k3), list(v)] if k3 is not None else [(k, k2), list(v)]
                    for k2, c2_dict in c_dict.items()
                    for k3, v in c2_dict.items()
                ]
                for k, c_dict in classifier_categories.items()
            ],
            [],
        )
    )

    return classifier_bow


def taxonomy_to_categories(taxonomy):
    categories = {
        (topic["filterable_title"], topic["title"]): {
            (subtopic["filterable_title"], subtopic["title"]): sum(
                [
                    [(subsubtopic["filterable_title"], subsubtopic["title"])]
                    for subsubtopic in subtopic["child_topics"]
                ],
                [],
            )
            if "child_topics" in subtopic
            else []
            for subtopic in topic["child_topics"]
        }
        for topic in taxonomy["topics"]
    }
    return categories



import numpy as np
import math
import re
from sortedcontainers import SortedDict
from nltk.corpus import stopwords, wordnet
from nltk.stem.wordnet import WordNetLemmatizer

from .utils import cosine_similarities
stopwords.ensure_loaded()
wordnet.ensure_loaded()

re_ws = re.compile(r"\s+")
re_num = re.compile(r"[^\w\s\']", flags=re.UNICODE)
DEFAULT_THRESHOLD = 0.1
DEFAULT_WEIGHTING = {"C": 2, "SC": 2, "SSC": 1, "WC": 3, "WSSC": 5}
EXTRA_STOPWORDS = {
    "english": ["statistics", "data", "measure", "measures"],
    "welsh": [],
}
DEFAULT_STOPWORDS_LANGUAGE = "english"


class WModel:
    def __init__(self, model):
        self.model = model
        self._model_dims = model.get_dims()

    def __getitem__(self, name):
        # Save Rust worrying about lifetime of a numpy array
        a = np.zeros((self._model_dims,), dtype=np.float32)

        self.model.load_embedding(name, a)

        return a


class Category:
    vector = None
    words = None

    def __init__(self, key, bow, model, weighting):
        self.key = key
        self.bow = bow
        self._weighting = weighting

        self._set_vector(model)
        self._set_words()

    def _set_vector(self, model):
        vector = np.mean([
            self._weighting[code] * model[w] for code, w in self.bow
        ], axis=0)
        self.vector = vector / sum(
            [self._weighting[code] for code, _ in self.bow]
        )

    def _set_words(self):
        self.words = [w for _, w in self.bow]


class CategoryManager:
    _stop_words = None
    _model = None
    _classifier_bow = None
    _topic_vectors = None

    def __init__(self, word_model, settings):
        self._categories = SortedDict()
        self._model = WModel(word_model)
        stopwords_language = settings.get(
            "STOPWORDS_LANGUAGE",
            DEFAULT_STOPWORDS_LANGUAGE
        )
        extra_stopwords = (
            settings
                .get("EXTRA_STOPWORDS", {})
                .get(stopwords_language, EXTRA_STOPWORDS[stopwords_language])
        )
        self._stop_words = (
            stopwords.words(stopwords_language) + extra_stopwords
        )
        self._significance = np.vectorize(
            self._significance_for_vector, signature="(m)->()"
        )
        self._ltzr = WordNetLemmatizer()
        self.all_words = {}
        self._weighting = settings.get("WEIGHTING", DEFAULT_WEIGHTING)
        self._threshold = settings.get("CATEGORY_THRESHOLD", DEFAULT_THRESHOLD)

    def set_all_words(self, all_words):
        total = sum(all_words.values())
        def scale(c):
            return 0.25 + math.exp(1000 * (1 - c) / total) * 0.75
        self.all_words = {w: scale(c) for w, c in all_words.items()}

    def _scale_by_frequency(self, word):
        lword = self._ltzr.lemmatize(word)
        if lword in self.all_words:
            scale = self.all_words[lword]
            return scale

        return 1.0

    def add_categories_from_bow(self, name, classifier_bow):
        self._categories[name] = SortedDict(
            (k, Category(k, bow, self._model, self._weighting))
            for k, bow in classifier_bow.items()
        )

    def closest(self, text, cat, classifier_bow_vec):
        word_list = set(sum(self.strip_document(text), []))
        word_scores = [
            (
                word,
                cosine_similarities(self._model[word], classifier_bow_vec[cat]).mean()
            )
            for word in word_list
            if cat in classifier_bow_vec
        ]
        return [
            word
            for word, score in sorted(
                word_scores, key=lambda word: word[1], reverse=True
            )
            if score > 0.3
        ]

    def strip_document(self, doc):
        if isinstance(doc, list):
            doc = " ".join(doc)

        docs = doc.split(",")
        word_list = []
        for doc in docs:
            doc = doc.replace("\n", " ").replace("_", " ").replace("'", "").lower()
            doc = re_ws.sub(" ", re_num.sub("", doc)).strip()

            if doc == "":
                return []

            word_list.append([w for w in doc.split(" ") if w not in self._stop_words])

        return word_list

    def test_category(self, sentence, category, category_group="dtcats"):
        cat = self._categories[category_group][category]

        clean = self.strip_document(sentence)

        if not clean:
            return []

        classifiers = {w: self._weighting[code] * self._model[w] for code, w in cat.bow}

        tags = {}
        for words in clean:
            if not words:
                continue

            vec = np.mean([self._model[w] for w in words], axis=0)
            result = cosine_similarities(vec, [cat.vector])[0]

            tags[" ".join(words)] = {
                "overall": result,
                "by-classifier": {
                    w: cosine_similarities(vec, [v])[0] for w, v in classifiers.items()
                },
            }

        return {
            "tags": tags,
            "vector": np.linalg.norm(cat.vector),
            "significance": self._significance_for_vector(cat.vector),
            "weightings": {w: self._weighting[code] for code, w in cat.bow},
        }

    @staticmethod
    def _significance_for_vector(vector):
        return min(max(0.5, np.linalg.norm(vector) * 1e2), 1.25)

    def test(self, sentence, category_group="dtcats"):
        categories = self._categories[category_group]

        clean = self.strip_document(sentence)

        if not clean:
            return []

        topic_vectors = np.array([c.vector for c in categories.values()])
        significance = self._significance(topic_vectors)

        tags = set()
        for words in clean:
            if not words:
                continue

            vec = np.mean(
                [self._model[w] * self._scale_by_frequency(w) for w in words], axis=0
            )
            result = cosine_similarities(vec, topic_vectors)
            result = np.multiply(result, significance)

            top = np.nonzero(result > self._threshold)[0]

            tags.update({(result[i], categories.keys()[i]) for i in top})

        return sorted(tags, reverse=True)
